import typing
from base64 import b64decode, b64encode
from itsdangerous import BadSignature, TimestampSigner
from starlette.datastructures import Secret

from ..serializers import JsonSerializer, Serializer
from .base import SessionBackend


class CookieBackend(SessionBackend):
    """Stores session data in the browser's cookie as a signed string."""

    def __init__(
        self,
        secret_key: typing.Union[str, Secret],
        max_age: int,
        serializer: Serializer = None,
    ):
        self._signer = TimestampSigner(str(secret_key))
        self._max_age = max_age
        self._serializer = serializer or JsonSerializer()

    async def read(self, session_id: str) -> typing.Dict:
        """A session_id is a signed session value."""
        try:
            data = self._signer.unsign(session_id, max_age=self._max_age)
            return self._serializer.deserialize(b64decode(data).decode())
        except BadSignature:
            return {}

    async def write(self, data: typing.Dict, session_id: typing.Optional[str] = None) -> str:
        """The data is a session id in this backend."""
        encoded_data = b64encode(self._serializer.serialize(data).encode("utf-8"))
        return self._signer.sign(encoded_data).decode("utf-8")

    async def remove(self, session_id: str) -> None:
        """Session data stored on client side - no way to remove it."""

    async def exists(self, session_id: str) -> bool:
        return False



import typing

from .backends.base import SessionBackend


class SessionError(Exception):
    """Base class for session exceptions."""


class SessionNotLoaded(SessionError):
    pass


class ImproperlyConfigured(SessionError):
    """Exception is raised when some settings are missing or misconfigured."""


class Session:
    def __init__(self, backend: SessionBackend, session_id: str = None) -> None:
        self.session_id = session_id
        self._data: typing.Dict[str, typing.Any] = {}
        self._backend = backend
        self.is_loaded = False
        self._is_modified = False

    @property
    def is_empty(self) -> bool:
        """Check if session has data."""
        return len(self.keys()) == 0

    @property
    def is_modified(self) -> bool:
        """Check if session data has been modified,"""
        return self._is_modified

    @property
    def data(self) -> typing.Dict:
        if not self.is_loaded:
            raise SessionNotLoaded("Session is not loaded.")
        return self._data

    @data.setter
    def data(self, value: typing.Dict[str, typing.Any]) -> None:
        self._data = value

    async def load(self) -> None:
        """Load data from the backend.
        Subsequent calls do not take any effect."""
        if self.is_loaded:
            return

        if not self.session_id:
            self.data = {}
        else:
            self.data = await self._backend.read(self.session_id)

        self.is_loaded = True

    async def persist(self) -> str:
        self.session_id = await self._backend.write(self.data, self.session_id)
        return self.session_id

    async def delete(self) -> None:
        if self.session_id:
            self.data = {}
            self._is_modified = True
            await self._backend.remove(self.session_id)

    async def flush(self) -> str:
        self._is_modified = True
        await self.delete()
        return await self.regenerate_id()

    async def regenerate_id(self) -> str:
        self.session_id = await self._backend.generate_id()
        self._is_modified = True
        return self.session_id

    def keys(self) -> typing.KeysView[str]:
        return self.data.keys()

    def values(self) -> typing.ValuesView[typing.Any]:
        return self.data.values()

    def items(self) -> typing.ItemsView[str, typing.Any]:
        return self.data.items()

    def pop(self, key: str, default: typing.Any = None) -> typing.Any:
        self._is_modified = True
        return self.data.pop(key, default)

    def get(self, name: str, default: typing.Any = None) -> typing.Any:
        return self.data.get(name, default)

    def setdefault(self, key: str, default: typing.Any) -> None:
        self._is_modified = True
        self.data.setdefault(key, default)

    def clear(self) -> None:
        self._is_modified = True
        self.data.clear()

    def update(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        self._is_modified = True
        self.data.update(*args, **kwargs)

    def __contains__(self, key: str) -> bool:
        return key in self.data

    def __setitem__(self, key: str, value: typing.Any) -> None:
        self._is_modified = True
        self.data[key] = value

    def __getitem__(self, key: str) -> typing.Any:
        return self.data[key]

    def __delitem__(self, key: str) -> None:
        self._is_modified = True
        del self.data[key]



import abc
import secrets
import typing


class SessionBackend(abc.ABC):
    """Base class for session backends."""

    @abc.abstractmethod
    async def read(self, session_id: str) -> typing.Dict[str, typing.Any]:  # pragma: no cover
        """Read session data from the storage."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def write(self, data: typing.Dict, session_id: typing.Optional[str] = None) -> str:  # pragma: no cover
        """Write session data to the storage."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def remove(self, session_id: str) -> None:  # pragma: no cover
        """Remove session data from the storage."""
        raise NotImplementedError()

    @abc.abstractmethod
    async def exists(self, session_id: str) -> bool:  # pragma: no cover
        """Test if storage contains session data for a given session_id."""
        raise NotImplementedError()

    async def generate_id(self) -> str:
        """Generate a new session id."""
        return secrets.token_hex(128)



import typing

from .base import SessionBackend


class InMemoryBackend(SessionBackend):
    """Stores session data in a dictionary."""

    def __init__(self) -> None:
        self.data: dict = {}

    async def read(self, session_id: str) -> typing.Dict:
        return self.data.get(session_id, {}).copy()

    async def write(self, data: typing.Dict, session_id: typing.Optional[str] = None) -> str:
        session_id = session_id or await self.generate_id()
        self.data[session_id] = data
        return session_id

    async def remove(self, session_id: str) -> None:
        del self.data[session_id]

    async def exists(self, session_id: str) -> bool:
        return session_id in self.data



from .backends import CookieBackend, InMemoryBackend, SessionBackend
from .middleware import SessionMiddleware
from .serializers import JsonSerializer, Serializer
from .session import ImproperlyConfigured, Session, SessionError, SessionNotLoaded

__all__ = [
    "SessionMiddleware",
    "Session",
    "SessionNotLoaded",
    "ImproperlyConfigured",
    "Serializer",
    "JsonSerializer",
    "SessionError",
    "SessionBackend",
    "InMemoryBackend",
    "CookieBackend",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['starsessions', 'starsessions.backends']

package_data = \
{'': ['*']}

install_requires = \
['itsdangerous>=2.0.1,<3.0.0', 'starlette>=0,<1']

extras_require = \
{'redis': ['aioredis>=2.0.0,<3.0.0']}

setup_kwargs = {
    'name': 'starsessions',
    'version': '1.2.3',
    'description': 'Pluggable session support for Starlette.',
    'long_description': '## Pluggable session support for Starlette and FastAPI frameworks\n\nThis package is based on this long standing [pull request](https://github.com/encode/starlette/pull/499) in the\nmainstream by the same author.\n\n## Installation\n\nInstall `starsessions` using PIP or poetry:\n\n```bash\npip install starsessions\n# or\npoetry add starsessions\n```\n\nUse `redis` extra for [Redis support](#redis).\n\n## Quick start\n\nSee example application in [`examples/`](examples) directory of this repository.\n\n## Enable session support\n\nIn order to enable session support add `starsessions.SessionMiddleware` to your application.\n\n```python\nfrom starlette.applications import Starlette\nfrom starlette.middleware import Middleware\nfrom starsessions import SessionMiddleware\n\nmiddleware = [\n    Middleware(SessionMiddleware, secret_key=\'TOP SECRET\'),\n]\n\napp = Starlette(middleware=middleware, **other_options)\n```\n\n### Session autoloading\n\nNote, for performance reasons session won\'t be autoloaded by default, you need to explicitly\ncall `await request.session.load()` before accessing the session otherwise `SessionNotLoaded` exception will be raised.\nYou can change this behavior by passing `autoload=True` to your middleware settings:\n\n```python\nMiddleware(SessionMiddleware, secret_key=\'TOP SECRET\', autoload=True)\n```\n\n### Cookie path\n\nYou can pass `path` arguments to enable session cookies on specific URLs. For example, to activate session cookie only\nfor admin area (which is hosted under `/admin` path prefix), use `path="/admin"` middleware argument.\n\n```python\nMiddleware(SessionMiddleware, path = \'/admin\', ...)\n```\n\nAll other URLs not matching value of `path` will not receive cookie thus session will be unavailable.\n\n### Cookie domain\n\nYou can also specify which hosts can receive a cookie by passing `domain` argument to the middleware.\n\n```python\nMiddleware(SessionMiddleware, domain = \'example.com\', ...)\n```\n\n> Note, this makes session cookie available for subdomains too.\n> For example, when you set `domain=example.com` then session cookie will be available on subdomains like `app.example.com`.\n\n### Session-only cookies\n\nIf you want session cookie to automatically remove from tbe browser when tab closes then set `max_age` to `0`:\n\n```python\nMiddleware(SessionMiddleware, max_age=0, ...)\n```\n\n### Default session backend\n\nThe default backend is `CookieBackend`. You don\'t need to configure it just pass `secret_key` argument and the backend\nwill be automatically configured for you.\n\n## Change default backend\n\nWhen you want to use a custom session storage then pass a desired backend instance via `backend` argument of the\nmiddleware.\n\n```python\nfrom starlette.applications import Starlette\nfrom starlette.middleware.sessions import SessionMiddleware\nfrom starlette.sessions import CookieBackend\n\nbackend = CookieBackend(secret_key=\'secret\', max_age=3600)\n\napp = Starlette()\napp.add_middleware(SessionMiddleware, backend=backend)\n```\n\n## Built-in backends\n\n### Memory\n\nClass: `starsessions.InMemoryBackend`\n\nSimply stores data in memory. The data is cleared after server restart. Mostly for use with unit tests.\n\n### CookieBackend\n\nClass: `starsessions.CookieBackend`\n\nStores session data in a signed cookie on the client. This is the default backend.\n\n### Redis\n\nClass: `starsessions.backends.redis.RedisBackend`\n\n> Requires [aioredis](https://aioredis.readthedocs.io/en/latest/getting-started/),\n> use `pip install starsessions[redis]` or `poetry add starsessions[redis]`\n\nStores session data in a Redis server. The backend accepts either connection URL or an instance of `aioredis.Redis`.\n\n```python\nimport aioredis\nfrom starsessions.backends.redis import RedisBackend\n\nbackend = RedisBackend(\'redis://localhost\')\n# or\nredis = aioredis.from_url(\'redis://localhost\')\n\nbackend = RedisBackend(connection=redis)\n```\n\n## Custom backend\n\nCreating new backends is quite simple. All you need is to extend `starsessions.SessionBackend`\nclass and implement abstract methods.\n\nHere is an example of how we can create a memory-based session backend. Note, it is important that `write` method\nreturns session ID as a string value.\n\n```python\nfrom starlette.sessions import SessionBackend\nfrom typing import Dict\n\n\n# instance of class which manages session persistence\n\nclass InMemoryBackend(SessionBackend):\n    def __init__(self):\n        self._storage = {}\n\n    async def read(self, session_id: str) -> Dict:\n        """ Read session data from a data source using session_id. """\n        return self._storage.get(session_id, {})\n\n    async def write(self, data: Dict, session_id: str = None) -> str:\n        """ Write session data into data source and return session id. """\n        session_id = session_id or await self.generate_id()\n        self._storage[session_id] = data\n        return session_id\n\n    async def remove(self, session_id: str):\n        """ Remove session data. """\n        del self._storage[session_id]\n\n    async def exists(self, session_id: str) -> bool:\n        return session_id in self._storage\n```\n\n## Serializers\n\nSometimes you cannot pass raw session data to the backend. The data must be serialized into something the backend can\nhandle.\n\nSome backends (like `RedisBackend`) optionally accept `serializer` argument that will be used to serialize and\ndeserialize session data. By default, we provide (and use) `starsessions.JsonSerializer` but you can implement your own\nby extending `starsessions.Serializer` class.\n',
    'author': 'alex.oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alex-oleshkevich/starsessions',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



import abc
import json
import typing as t


class Serializer(abc.ABC):  # pragma: no cover
    @abc.abstractmethod
    def serialize(self, data: t.Any) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def deserialize(self, data: str) -> t.Dict:
        raise NotImplementedError()


class JsonSerializer(Serializer):
    def __init__(
        self,
        json_encoder: t.Type[json.JSONEncoder] = json.JSONEncoder,
        json_decoder: t.Type[json.JSONDecoder] = json.JSONDecoder,
    ) -> None:
        self._json_encoder = json_encoder
        self._json_decoder = json_decoder

    def serialize(self, data: t.Any) -> str:
        return json.dumps(data, cls=self._json_encoder)

    def deserialize(self, data: str) -> t.Dict:
        return json.loads(data, cls=self._json_decoder)
